@@ -75,10 +75,13 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	struct vpn_option *old_dtls_opts = vpninfo->dtls_options;
 	const char *old_addr = vpninfo->vpn_addr;
 	const char *old_netmask = vpninfo->vpn_netmask;
+	const char *old_addr6 = vpninfo->vpn_addr6;
+	const char *old_netmask6 = vpninfo->vpn_netmask6;
 	struct split_include *inc;
 
 	/* Clear old options which will be overwritten */
 	vpninfo->vpn_addr = vpninfo->vpn_netmask = NULL;
+	vpninfo->vpn_addr6 = vpninfo->vpn_netmask6 = NULL;
 	vpninfo->cstp_options = vpninfo->dtls_options = NULL;
 	vpninfo->vpn_domain = vpninfo->vpn_proxy_pac = NULL;
 
@@ -106,9 +109,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	if (vpninfo->deflate)
 		openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-Accept-Encoding: deflate;q=1.0\r\n");
 	openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-MTU: %d\r\n", vpninfo->mtu);
-	/* To enable IPv6, send 'IPv6,IPv4'.
-	   We don't know how most of that works yet though. */
-	openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-Address-Type: IPv4\r\n");
+	openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-Address-Type: %s\r\n",
+			       vpninfo->disable_ipv6?"IPv4":"IPv6,IPv4");
 	openconnect_SSL_printf(vpninfo->https_ssl, "X-DTLS-Master-Secret: ");
 	for (i = 0; i < sizeof(vpninfo->dtls_secret); i++)
 		openconnect_SSL_printf(vpninfo->https_ssl, "%02X", vpninfo->dtls_secret[i]);
@@ -217,9 +219,15 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		} else if (!strcmp(buf + 7, "MTU")) {
 			vpninfo->mtu = atol(colon);
 		} else if (!strcmp(buf + 7, "Address")) {
-			vpninfo->vpn_addr = new_option->value;
+			if (strchr(new_option->value, ':'))
+				vpninfo->vpn_addr6 = new_option->value;
+			else
+				vpninfo->vpn_addr = new_option->value;
 		} else if (!strcmp(buf + 7, "Netmask")) {
-			vpninfo->vpn_netmask = new_option->value;
+			if (strchr(new_option->value, ':'))
+				vpninfo->vpn_netmask6 = new_option->value;
+			else
+				vpninfo->vpn_netmask = new_option->value;
 		} else if (!strcmp(buf + 7, "DNS")) {
 			int j;
 			for (j = 0; j < 3; j++) {
@@ -257,26 +265,38 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		}
 	}
 
-	if (!vpninfo->vpn_addr) {
+	if (!vpninfo->vpn_addr && !vpninfo->vpn_addr6) {
 		vpninfo->progress(vpninfo, PRG_ERR, "No IP address received. Aborting\n");
 		return -EINVAL;
 	}
-	if (!vpninfo->vpn_netmask)
-		vpninfo->vpn_netmask = "255.255.255.255";
 	if (old_addr) {
 		if (strcmp(old_addr, vpninfo->vpn_addr)) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Reconnect gave different IP address (%s != %s)\n",
+			vpninfo->progress(vpninfo, PRG_ERR, "Reconnect gave different Legacy IP address (%s != %s)\n",
 				vpninfo->vpn_addr, old_addr);
 			return -EINVAL;
 		}
 	}
 	if (old_netmask) {
 		if (strcmp(old_netmask, vpninfo->vpn_netmask)) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Reconnect gave different netmask (%s != %s)\n",
+			vpninfo->progress(vpninfo, PRG_ERR, "Reconnect gave different Legacy IP netmask (%s != %s)\n",
 				vpninfo->vpn_netmask, old_netmask);
 			return -EINVAL;
 		}
 	}
+	if (old_addr6) {
+		if (strcmp(old_addr6, vpninfo->vpn_addr6)) {
+			vpninfo->progress(vpninfo, PRG_ERR, "Reconnect gave different IPv6 address (%s != %s)\n",
+				vpninfo->vpn_addr6, old_addr6);
+			return -EINVAL;
+		}
+	}
+	if (old_netmask6) {
+		if (strcmp(old_netmask6, vpninfo->vpn_netmask6)) {
+			vpninfo->progress(vpninfo, PRG_ERR, "Reconnect gave different IPv6 netmask (%s != %s)\n",
+				vpninfo->vpn_netmask6, old_netmask6);
+			return -EINVAL;
+		}
+	}
 
 	while (old_dtls_opts) {
 		struct vpn_option *tmp = old_dtls_opts;
@@ -369,14 +389,13 @@ static int cstp_reconnect(struct openconnect_info *vpninfo)
 	return 0;
 }
 
-static int inflate_and_queue_packet(struct openconnect_info *vpninfo, int type, void *buf, int len)
+static int inflate_and_queue_packet(struct openconnect_info *vpninfo, void *buf, int len)
 {
 	struct pkt *new = malloc(sizeof(struct pkt) + vpninfo->mtu);
 
 	if (!new)
 		return -ENOMEM;
 
-	new->type = type;
 	new->next = NULL;
 
 	vpninfo->inflate_strm.next_in = buf;
@@ -461,7 +480,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			vpninfo->progress(vpninfo, PRG_TRACE,
 					  "Received uncompressed data packet of %d bytes\n",
 					  payload_len);
-			queue_new_packet(&vpninfo->incoming_queue, AF_INET, buf + 8,
+			queue_new_packet(&vpninfo->incoming_queue, buf + 8,
 					 payload_len);
 			work_done = 1;
 			continue;
@@ -483,7 +502,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				vpninfo->progress(vpninfo, PRG_ERR, "Compressed packet received in !deflate mode\n");
 				goto unknown_pkt;
 			}
-			inflate_and_queue_packet(vpninfo, AF_INET, buf + 8, payload_len);
+			inflate_and_queue_packet(vpninfo, buf + 8, payload_len);
 			work_done = 1;
 			continue;
 
@@ -502,6 +521,14 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		return 1;
 	}
 
+	ret = SSL_get_error(vpninfo->https_ssl, len);
+	if (ret == SSL_ERROR_SYSCALL || ret == SSL_ERROR_ZERO_RETURN) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "SSL read error %d (server probably closed connection); reconnecting.\n",
+				  ret);
+			goto do_reconnect;
+	}
+
 
 	/* If SSL_write() fails we are expected to try again. With exactly
 	   the same data, at exactly the same location. So we keep the
@@ -609,10 +636,6 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		vpninfo->outgoing_queue = this->next;
 		vpninfo->outgoing_qlen--;
 
-		/* FIXME: Don't know how to handle IPv6 yet */
-		if (this->type != AF_INET)
-			continue;
-
 		if (vpninfo->deflate) {
 			unsigned char *adler;
 			int ret;
@@ -678,18 +701,19 @@ int cstp_bye(struct openconnect_info *vpninfo, char *reason)
 		return 0;
 
 	reason_len = strlen(reason);
-	bye_pkt = malloc(reason_len + 8);
+	bye_pkt = malloc(reason_len + 9);
 	if (!bye_pkt)
 		return -ENOMEM;
 
 	memcpy(bye_pkt, data_hdr, 8);
-	memcpy(bye_pkt + 8, reason, reason_len);
+	memcpy(bye_pkt + 9, reason, reason_len);
 
-	bye_pkt[4] = reason_len >> 8;
-	bye_pkt[5] = reason_len & 0xff;
+	bye_pkt[4] = (reason_len + 1) >> 8;
+	bye_pkt[5] = (reason_len + 1) & 0xff;
 	bye_pkt[6] = AC_PKT_DISCONN;
+	bye_pkt[8] = 0xb0;
 
-	SSL_write(vpninfo->https_ssl, bye_pkt, reason_len + 8);
+	SSL_write(vpninfo->https_ssl, bye_pkt, reason_len + 9);
 	free(bye_pkt);
 
 	vpninfo->progress(vpninfo, PRG_INFO,
