@@ -106,7 +106,7 @@ static unsigned char hex(const char *data)
 int connect_dtls_socket(struct openconnect_info *vpninfo)
 {
 	STACK_OF(SSL_CIPHER) *ciphers;
-	SSL_METHOD *dtls_method;
+	method_const SSL_METHOD *dtls_method;
 	SSL_CIPHER *dtls_cipher;
 	SSL *dtls_ssl;
 	BIO *dtls_bio;
@@ -211,6 +211,7 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 	SSL_set_bio(dtls_ssl, dtls_bio, dtls_bio);
 
 #ifndef SSL_OP_CISCO_ANYCONNECT
+#warning Your version of OpenSSL does not seem to support Cisco DTLS compatibility
 #define SSL_OP_CISCO_ANYCONNECT 0x8000
 #endif
 	SSL_set_options(dtls_ssl, SSL_OP_CISCO_ANYCONNECT);
@@ -390,7 +391,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 		switch(buf[0]) {
 		case AC_PKT_DATA:
-			queue_new_packet(&vpninfo->incoming_queue, AF_INET, buf+1, len-1);
+			queue_new_packet(&vpninfo->incoming_queue, buf+1, len-1);
 			work_done = 1;
 			break;
 
@@ -483,10 +484,6 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		vpninfo->outgoing_queue = this->next;
 		vpninfo->outgoing_qlen--;
 
-		/* FIXME: Don't know how to handle IPv6 yet */
-		if (this->type != AF_INET)
-			continue;
-
 		/* One byte of header */
 		this->hdr[7] = AC_PKT_DATA;
 
