@@ -24,8 +24,10 @@
  */
 
 #include <stdio.h>
+#include <syslog.h>
 #include <stdarg.h>
 #include <stdlib.h>
+#include <signal.h>
 #include <string.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -84,6 +86,9 @@ static struct option long_options[] = {
 	{"authgroup", 1, 0, '9'},
 	{"servercert", 1, 0, 0x01},
 	{"key-password-from-fsid", 0, 0, 0x02},
+	{"useragent", 1, 0, 0x03},
+	{"csd-user", 1, 0, 0x04},
+	{"disable-ipv6", 0, 0, 0x05},
 	{NULL, 0, 0, 0},
 };
 
@@ -104,6 +109,7 @@ void usage(void)
 	printf("  -i, --interface=IFNAME          Use IFNAME for tunnel interface\n");
 	printf("  -l, --syslog                    Use syslog for progress messages\n");
 	printf("  -U, --setuid=USER               Drop privileges after connecting\n");
+	printf("      --csd-user=USER           Drop privileges during CSD execution\n");
 	printf("  -m, --mtu=MTU                   Request MTU from server\n");
 	printf("  -p, --key-password=PASS         Set key passphrase or TPM SRK PIN\n");
 	printf("      --key-password-from-fsid    Key passphrase is fsid of file system\n");
@@ -119,12 +125,14 @@ void usage(void)
 	printf("      --cookieonly                Fetch webvpn cookie only; don't connect\n");
 	printf("      --printcookie               Print webvpn cookie before connecting\n");
 	printf("      --cafile=FILE               Cert file for server verification\n");
+	printf("      --disable-ipv6              Do not ask for IPv6 connectivity\n");
 	printf("      --dtls-ciphers=LIST         OpenSSL ciphers to support for DTLS\n");
 	printf("      --no-dtls                   Disable DTLS\n");
 	printf("      --no-passwd                 Disable password/SecurID authentication\n");
 	printf("      --passwd-on-stdin           Read password from standard input\n");
 	printf("      --reconnect-timeout         Connection retry timeout in seconds\n");
 	printf("      --servercert                Server's certificate SHA1 fingerprint\n");
+	printf("      --useragent=STRING          HTTP header User-Agent: field\n");
 	exit(1);
 }
 
@@ -146,11 +154,19 @@ static void read_stdin(char **string)
 	if (c)
 		*c = 0;
 }
+static void handle_sigusr(int sig)
+{
+	if (sig == SIGUSR1)
+		verbose = PRG_TRACE;
+	else if (sig == SIGUSR2)
+		verbose = PRG_INFO;
+}
 
 int main(int argc, char **argv)
 {
 	struct openconnect_info *vpninfo;
 	struct utsname utsbuf;
+	struct sigaction sa;
 	int cookieonly = 0;
 	int use_syslog = 0;
 	uid_t uid = getuid();
@@ -167,13 +183,15 @@ int main(int argc, char **argv)
 
 	/* Set up some defaults */
 	vpninfo->tun_fd = vpninfo->ssl_fd = vpninfo->dtls_fd = vpninfo->new_dtls_fd = -1;
-	vpninfo->useragent = openconnect_create_useragent("Open AnyConnect VPN Agent %s");
+	vpninfo->useragent = openconnect_create_useragent("Open AnyConnect VPN Agent");
 	vpninfo->mtu = 1406;
 	vpninfo->deflate = 1;
 	vpninfo->dtls_attempt_period = 60;
 	vpninfo->max_qlen = 10;
 	vpninfo->reconnect_interval = RECONNECT_INTERVAL_MIN;
 	vpninfo->reconnect_timeout = 300;
+	vpninfo->uid_csd = 0;
+	vpninfo->uid_csd_given = 0;
 
 	if (RAND_bytes(vpninfo->dtls_secret, sizeof(vpninfo->dtls_secret)) != 1) {
 		fprintf(stderr, "Failed to initialise DTLS secret\n");
@@ -303,6 +321,24 @@ int main(int argc, char **argv)
 			}
 			break;
 		}
+		case 0x04: {
+			char *strend;
+			vpninfo->uid_csd = strtol(optarg, &strend, 0);
+			if (strend[0]) {
+				struct passwd *pw = getpwnam(optarg);
+				if (!pw) {
+					fprintf(stderr, "Invalid user \"%s\"\n",
+						optarg);
+					exit(1);
+				}
+				vpninfo->uid_csd = pw->pw_uid;
+			}
+			vpninfo->uid_csd_given = 1;
+			break;
+		}
+		case 0x05:
+			vpninfo->disable_ipv6 = 1;
+			break;
 		case 'Q':
 			vpninfo->max_qlen = atol(optarg);
 			if (!vpninfo->max_qlen) {
@@ -326,6 +362,10 @@ int main(int argc, char **argv)
 		case 0x02:
 			do_passphrase_from_fsid = 1;
 			break;
+		case 0x03:
+			free(vpninfo->useragent);
+			vpninfo->useragent = optarg;
+			break;
 		default:
 			usage();
 		}
@@ -344,6 +384,11 @@ int main(int argc, char **argv)
 	} else {
 		vpninfo->progress = write_progress;
 	}
+	memset(&sa, 0, sizeof(sa));
+	sa.sa_handler = handle_sigusr;
+
+	sigaction(SIGUSR1, &sa, NULL);
+	sigaction(SIGUSR2, &sa, NULL);
 
 	if (vpninfo->sslkey && do_passphrase_from_fsid)
 		passphrase_from_fsid(vpninfo);
@@ -390,8 +435,10 @@ int main(int argc, char **argv)
 		fprintf(stderr, "Set up DTLS failed; using SSL instead\n");
 
 	vpninfo->progress(vpninfo, PRG_INFO,
-			  "Connected %s as %s, using %s\n", vpninfo->ifname,
-			  vpninfo->vpn_addr,
+			  "Connected %s as %s%s%s, using %s\n", vpninfo->ifname,
+			  vpninfo->vpn_addr?:"",
+			  (vpninfo->vpn_addr6 && vpninfo->vpn_addr)?" + ":"",
+			  vpninfo->vpn_addr6?:"",
 			  (vpninfo->dtls_fd == -1) ?
 			      (vpninfo->deflate ? "SSL + deflate" : "SSL")
 			      : "DTLS");
@@ -412,16 +459,25 @@ int main(int argc, char **argv)
 static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buflen)
 {
 	int config_fd;
+	int err;
 
 	config_fd = open(vpninfo->xmlconfig, O_WRONLY|O_TRUNC|O_CREAT, 0644);
-	if (!config_fd) {
+	if (config_fd < 0) {
+		err = errno;
 		fprintf(stderr, "Failed to open %s for write: %s\n",
-			vpninfo->xmlconfig, strerror(errno));
-		return -errno;
+			vpninfo->xmlconfig, strerror(err));
+		return -err;
 	}
 
 	/* FIXME: We should actually write to a new tempfile, then rename */
-	write(config_fd, buf, buflen);
+	if(write(config_fd, buf, buflen) != buflen) {
+		err = errno;
+		fprintf(stderr, "Failed to write config to %s: %s\n",
+			vpninfo->xmlconfig, strerror(err));
+
+		return -err;
+	}
+	  
 	return 0;
 }
 
