@@ -44,13 +44,12 @@ void queue_packet(struct pkt **q, struct pkt *new)
 	*q = new;
 }
 
-int queue_new_packet(struct pkt **q, int type, void *buf, int len)
+int queue_new_packet(struct pkt **q, void *buf, int len)
 {
 	struct pkt *new = malloc(sizeof(struct pkt) + len);
 	if (!new)
 		return -ENOMEM;
 
-	new->type = type;
 	new->len = len;
 	new->next = NULL;
 	memcpy(new->data, buf, len);
@@ -71,6 +70,7 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 	memset(&sa, 0, sizeof(sa));
 	sa.sa_handler = handle_sigint;
 
+	sigaction(SIGTERM, &sa, NULL);
 	sigaction(SIGINT, &sa, NULL);
 	sigaction(SIGHUP, &sa, NULL);
 
@@ -126,30 +126,13 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 
 		tv.tv_sec = timeout / 1000;
 		tv.tv_usec = (timeout % 1000) * 1000;
-		if (select(vpninfo->select_nfds, &rfds, &wfds, &efds, &tv) >= 0
-		    && FD_ISSET(vpninfo->ssl_fd, &rfds)) {
-			char b;
-			if (recv(vpninfo->ssl_fd, &b, 1, MSG_PEEK) == 0) {
-				/* Zero indicates EOF, which OpenSSL won't handle on read() */
-				vpninfo->progress(vpninfo, PRG_ERR, "Server closed connection!\n");
-				exit(1);
-			}
-		}
+
+		select(vpninfo->select_nfds, &rfds, &wfds, &efds, &tv);
 	}
 
 	cstp_bye(vpninfo, vpninfo->quit_reason);
 
-	if (vpninfo->vpnc_script) {
-		if (vpninfo->script_tun) {
-			kill(vpninfo->script_tun, SIGHUP);
-			close(vpninfo->tun_fd);
-		} else {
-			setenv("TUNDEV", vpninfo->ifname, 1);
-			setenv("reason", "disconnect", 1);
-			system(vpninfo->vpnc_script);
-		}
-	}
-
+	shutdown_tun(vpninfo);
 	return 0;
 }
 
