@@ -85,7 +85,6 @@ struct oc_auth_form {
 /****************************************************************************/
 
 struct pkt {
-	int type;
 	int len;
 	struct pkt *next;
 	unsigned char hdr[8];
@@ -130,6 +129,15 @@ struct split_include {
 struct openconnect_info {
 	char *redirect_url;
 
+	char *csd_token;
+	char *csd_ticket;
+	char *csd_stuburl;
+	char *csd_starturl;
+	char *csd_waiturl;
+	char *csd_preurl;
+
+	char *csd_scriptname;
+
 	char *vpn_name;
 
 	char sid_tokencode[9];
@@ -151,6 +159,8 @@ struct openconnect_info {
 	char *authgroup;
 	int nopasswd;
 	char *dtls_ciphers;
+	uid_t uid_csd;
+	int uid_csd_given;
 
 	char *cookie;
 	struct vpn_option *cookies;
@@ -169,6 +179,7 @@ struct openconnect_info {
 	z_stream deflate_strm;
 	uint32_t deflate_adler32;
 
+	int disable_ipv6;
 	int reconnect_timeout;
 	int reconnect_interval;
 	int dtls_attempt_period;
@@ -189,6 +200,8 @@ struct openconnect_info {
 	int mtu;
 	const char *vpn_addr;
 	const char *vpn_netmask;
+	const char *vpn_addr6;
+	const char *vpn_netmask6;
 	const char *vpn_dns[3];
 	const char *vpn_nbns[3];
 	const char *vpn_domain;
@@ -201,6 +214,10 @@ struct openconnect_info {
 	fd_set select_wfds;
 	fd_set select_efds;
 
+#ifdef __sun__
+	int ip_fd;
+	int tun_muxid;
+#endif
 	int tun_fd;
 	int ssl_fd;
 	int dtls_fd;
@@ -215,7 +232,7 @@ struct openconnect_info {
 	struct sockaddr *peer_addr;
 
 	int deflate;
-	const char *useragent;
+	char *useragent;
 
 	char *quit_reason;
 
@@ -242,12 +259,19 @@ struct openconnect_info {
 #define AC_PKT_COMPRESSED	8	/* Compressed data */
 #define AC_PKT_TERM_SERVER	9	/* Server kick */
 
+/* Ick */
+#if OPENSSL_VERSION_NUMBER >= 0x10000000L
+#define method_const const
+#else
+#define method_const
+#endif
 
 /****************************************************************************/
 
 /* tun.c */
 int setup_tun(struct openconnect_info *vpninfo);
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
+void shutdown_tun(struct openconnect_info *vpninfo);
 
 /* dtls.c */
 int setup_dtls(struct openconnect_info *vpninfo);
@@ -267,14 +291,17 @@ int  __attribute__ ((format (printf, 2, 3)))
 int openconnect_SSL_gets(SSL *ssl, char *buf, size_t len);
 int openconnect_open_https(struct openconnect_info *vpninfo);
 void openconnect_close_https(struct openconnect_info *vpninfo);
-int get_cert_fingerprint(X509 *cert, char *buf);
+int get_cert_md5_fingerprint(struct openconnect_info *vpninfo, X509 *cert,
+			     char *buf);
+int get_cert_sha1_fingerprint(struct openconnect_info *vpninfo, X509 *cert,
+			      char *buf);
 void report_ssl_errors(struct openconnect_info *vpninfo);
 int passphrase_from_fsid(struct openconnect_info *vpninfo);
 
 /* mainloop.c */
 int vpn_add_pollfd(struct openconnect_info *vpninfo, int fd, short events);
 int vpn_mainloop(struct openconnect_info *vpninfo);
-int queue_new_packet(struct pkt **q, int type, void *buf, int len);
+int queue_new_packet(struct pkt **q, void *buf, int len);
 void queue_packet(struct pkt **q, struct pkt *new);
 int keepalive_action(struct keepalive_info *ka, int *timeout);
 int ka_stalled_dpd_time(struct keepalive_info *ka, int *timeout);
@@ -286,7 +313,8 @@ int config_lookup_host(struct openconnect_info *vpninfo, const char *host);
 
 /* auth.c */
 int parse_xml_response(struct openconnect_info *vpninfo, char *response,
-		       char *request_body, int req_len);
+		       char *request_body, int req_len, char **method,
+		       char **request_body_type);
 
 /* http.c */
 int openconnect_obtain_cookie(struct openconnect_info *vpninfo);
