@@ -22,6 +22,7 @@
  *   Boston, MA 02110-1301 USA
  */
 
+#define _GNU_SOURCE
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netdb.h>
@@ -29,11 +30,14 @@
 #include <fcntl.h>
 #include <string.h>
 #include <ctype.h>
+#include <stdio.h>
 #if defined(__linux__)
 #include <sys/vfs.h>
 #elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
 #include <sys/param.h>
 #include <sys/mount.h>
+#elif defined (__sun__)
+#include <sys/statvfs.h>
 #endif
 
 #include <openssl/ssl.h>
@@ -249,8 +253,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		f = fopen(vpninfo->cert, "r");
 		if (!f) {
 			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Failed to open certificate file %s\n",
-					  vpninfo->cert);
+					  "Failed to open certificate file %s: %s\n",
+					  vpninfo->cert, strerror(errno));
 			return -ENOENT;
 		}
 		p12 = d2i_PKCS12_fp(f, NULL);
@@ -272,7 +276,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	if (!SSL_CTX_use_certificate_chain_file(vpninfo->https_ctx,
 						vpninfo->cert)) {
 		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Load certificate failed\n");
+				  "Loading certificate failed\n");
 		report_ssl_errors(vpninfo);
 		return -EINVAL;
 	}
@@ -283,8 +287,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 		if (!f) {
 			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Failed to open certificate file %s\n",
-					  vpninfo->cert);
+					  "Failed to open private key file %s: %s\n",
+					  vpninfo->cert, strerror(errno));
 			return -ENOENT;
 		}
 
@@ -294,7 +298,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 				vpninfo->cert_type = CERT_TYPE_TPM;
 				break;
 			} else if (!strcmp(buf, "-----BEGIN RSA PRIVATE KEY-----\n") ||
-				   !strcmp(buf, "-----BEGIN DSA PRIVATE KEY-----\n")) {
+				   !strcmp(buf, "-----BEGIN DSA PRIVATE KEY-----\n") ||
+				   !strcmp(buf, "-----BEGIN ENCRYPTED PRIVATE KEY-----\n")) {
 				vpninfo->cert_type = CERT_TYPE_PEM;
 				break;
 			}
@@ -323,7 +328,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 					    SSL_FILETYPE_PEM)) {
 		unsigned long err = ERR_peek_error();
 		
-		vpninfo->progress(vpninfo, PRG_ERR, "Private key failed\n");
+		vpninfo->progress(vpninfo, PRG_ERR, "Loading private key failed\n");
 		report_ssl_errors(vpninfo);
 
 #ifndef EVP_F_EVP_DECRYPTFINAL_EX
@@ -340,26 +345,56 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	return 0;
 }
 
-int get_cert_fingerprint(X509 *cert, char *buf)
+enum cert_hash_type {
+	EVP_MD5,
+	EVP_SHA1
+};
+
+static int get_cert_fingerprint(struct openconnect_info *vpninfo,
+				X509 *cert, enum cert_hash_type hash,
+				char *buf)
 {
 	unsigned char md[EVP_MAX_MD_SIZE];
 	unsigned int i, n;
 
-	if (!X509_digest(cert, EVP_sha1(), md, &n))
-		return -ENOMEM;
+	switch (hash) {
+	case EVP_MD5:
+		if (!X509_digest(cert, EVP_md5(), md, &n))
+			return -ENOMEM;
+		break;
+	case EVP_SHA1:
+		if (!X509_digest(cert, EVP_sha1(), md, &n))
+			return -ENOMEM;
+		break;
+	default:
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Unsupported SSL certificate hash function type\n");
+	}
 
 	for (i=0; i < n; i++) {
-		sprintf(&buf[i*2], "%02x", md[i]);
+		sprintf(&buf[i*2], "%02X", md[i]);
 	}
 	return 0;
 }
 
+int get_cert_md5_fingerprint(struct openconnect_info *vpninfo,
+			     X509 *cert, char *buf)
+{
+	return get_cert_fingerprint(vpninfo, cert, EVP_MD5, buf);
+}
+
+int get_cert_sha1_fingerprint(struct openconnect_info *vpninfo,
+			      X509 *cert, char *buf)
+{
+	return get_cert_fingerprint(vpninfo, cert, EVP_SHA1, buf);
+}
+
 static int check_server_cert(struct openconnect_info *vpninfo, X509 *cert)
 {
 	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
 	int ret;
 
-	ret = get_cert_fingerprint(cert, fingerprint);
+	ret = get_cert_sha1_fingerprint(vpninfo, cert, fingerprint);
 	if (ret)
 		return ret;
 
@@ -436,56 +471,72 @@ void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
-	SSL_METHOD *ssl3_method;
+	method_const SSL_METHOD *ssl3_method;
 	SSL *https_ssl;
 	BIO *https_bio;
 	int ssl_sock = -1;
 	int err;
-	struct addrinfo hints, *result, *rp;
 
-	memset(&hints, 0, sizeof(struct addrinfo));
-	hints.ai_family = AF_UNSPEC;
-	hints.ai_socktype = SOCK_STREAM;
-	hints.ai_flags = AI_PASSIVE;
-	hints.ai_protocol = 0;
-	hints.ai_canonname = NULL;
-	hints.ai_addr = NULL;
-	hints.ai_next = NULL;
-
-	err = getaddrinfo(vpninfo->hostname, "https", &hints, &result);
-	if (err) {
-		vpninfo->progress(vpninfo, PRG_ERR, "getaddrinfo failed: %s\n", gai_strerror(err));
-		return -EINVAL;
-	}
-
-	vpninfo->progress(vpninfo, PRG_INFO,
-			  "Attempting to connect to %s\n", vpninfo->hostname);
-
-	for (rp = result; rp ; rp = rp->ai_next) {
-		ssl_sock = socket(rp->ai_family, rp->ai_socktype,
-				  rp->ai_protocol);
-		if (ssl_sock < 0)
-			continue;
-		if (connect(ssl_sock, rp->ai_addr, rp->ai_addrlen) >= 0) {
-			/* Store the peer address we actually used, so that DTLS can
-			   use it again later */
-			vpninfo->peer_addr = malloc(rp->ai_addrlen);
-			if (!vpninfo->peer_addr) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Failed to allocate sockaddr storage\n");
-				close(ssl_sock);
-				return -ENOMEM;
-			}
-			vpninfo->peer_addrlen = rp->ai_addrlen;
-			memcpy(vpninfo->peer_addr, rp->ai_addr, rp->ai_addrlen);
-			break;
+	if (vpninfo->peer_addr) {
+		ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM, IPPROTO_IP);
+		if (ssl_sock < 0) {
+		reconn_err:
+			vpninfo->progress(vpninfo, PRG_ERR, "Failed to reconnect to host %s\n", vpninfo->hostname);
+			return -EINVAL;
 		}
-		close(ssl_sock);
-	}
-	freeaddrinfo(result);
+		if (connect(ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen))
+			goto reconn_err;
 
-	if (ssl_sock < 0) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to connect to host %s\n", vpninfo->hostname);
-		return -EINVAL;
+		
+		
+	} else {
+		struct addrinfo hints, *result, *rp;
+
+		memset(&hints, 0, sizeof(struct addrinfo));
+		hints.ai_family = AF_UNSPEC;
+		hints.ai_socktype = SOCK_STREAM;
+		hints.ai_flags = AI_PASSIVE | AI_NUMERICSERV;
+		hints.ai_protocol = 0;
+		hints.ai_canonname = NULL;
+		hints.ai_addr = NULL;
+		hints.ai_next = NULL;
+
+		err = getaddrinfo(vpninfo->hostname, "443", &hints, &result);
+		if (err) {
+			vpninfo->progress(vpninfo, PRG_ERR, "getaddrinfo failed: %s\n", gai_strerror(err));
+			return -EINVAL;
+		}
+
+		vpninfo->progress(vpninfo, PRG_INFO,
+				  "Attempting to connect to %s\n", vpninfo->hostname);
+
+		for (rp = result; rp ; rp = rp->ai_next) {
+			ssl_sock = socket(rp->ai_family, rp->ai_socktype,
+					  rp->ai_protocol);
+			if (ssl_sock < 0)
+				continue;
+			if (connect(ssl_sock, rp->ai_addr, rp->ai_addrlen) >= 0) {
+				/* Store the peer address we actually used, so that DTLS can
+				   use it again later */
+				vpninfo->peer_addr = malloc(rp->ai_addrlen);
+				if (!vpninfo->peer_addr) {
+					vpninfo->progress(vpninfo, PRG_ERR, "Failed to allocate sockaddr storage\n");
+					close(ssl_sock);
+					return -ENOMEM;
+				}
+				vpninfo->peer_addrlen = rp->ai_addrlen;
+				memcpy(vpninfo->peer_addr, rp->ai_addr, rp->ai_addrlen);
+				break;
+			}
+			close(ssl_sock);
+			ssl_sock = -1;
+		}
+		freeaddrinfo(result);
+		
+		if (ssl_sock < 0) {
+			vpninfo->progress(vpninfo, PRG_ERR, "Failed to connect to host %s\n", vpninfo->hostname);
+			return -EINVAL;
+		}
 	}
 	fcntl(ssl_sock, F_SETFD, FD_CLOEXEC);
 
@@ -562,22 +613,36 @@ void openconnect_init_openssl(void)
 	OpenSSL_add_all_algorithms ();
 }
 
+#ifdef __sun__
+int passphrase_from_fsid(struct openconnect_info *vpninfo)
+{
+	struct statvfs buf;
+
+	if (statvfs(vpninfo->sslkey, &buf)) {
+		int err = errno;
+		vpninfo->progress(vpninfo, PRG_ERR, "statvfs: %s\n", strerror(errno));
+		return -err;
+	}
+	if (asprintf(&vpninfo->cert_password, "%lx", buf.f_fsid))
+		return -ENOMEM;
+	return 0;
+}
+#else
 int passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
 	struct statfs buf;
 	unsigned *fsid = (unsigned *)&buf.f_fsid;
 	unsigned long long fsid64;
 
-	vpninfo->cert_password = malloc(17);
-	if (!vpninfo->cert_password)
-		return -ENOMEM;
-
 	if (statfs(vpninfo->sslkey, &buf)) {
 		int err = errno;
 		vpninfo->progress(vpninfo, PRG_ERR, "statfs: %s\n", strerror(errno));
 		return -err;
 	}
 	fsid64 = ((unsigned long long)fsid[0] << 32) | fsid[1];
-	sprintf(vpninfo->cert_password, "%llx", fsid64);
+
+	if (asprintf(&vpninfo->cert_password, "%llx", fsid64))
+		return -ENOMEM;
 	return 0;
 }
+#endif
