@@ -27,9 +27,7 @@
 #include <sys/socket.h>
 #include <sys/ioctl.h>
 #include <string.h>
-#ifdef __linux__
-#include <linux/if_tun.h>
-#endif
+#include <signal.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <netinet/in_systm.h>
@@ -38,13 +36,49 @@
 #include <net/if.h>
 #include <arpa/inet.h>
 #include <errno.h>
+#if defined(__sun__)
+#include <net/if_tun.h>
+#include <stropts.h>
+#include <sys/sockio.h>
+#endif
 
 #include "openconnect.h"
 
-#ifdef BSD
+/*
+ * If an if_tun.h include file was found anywhere (by the Makefile), it's 
+ * included. Else, we end up assuming that we have BSD-style devices such
+ * as /dev/tun0 etc.
+ */
+#ifdef IF_TUN_HDR
+#include IF_TUN_HDR
+#endif
+
+/*
+ * The OS X tun/tap driver doesn't provide a header file; you're expected
+ * to define this for yourself.
+ */
+#ifdef __APPLE__
+#define TUNSIFHEAD  _IOW('t', 96, int)
+#endif
+
+/*
+ * OpenBSD always puts the protocol family prefix onto packets. Other
+ * systems let us enable that with the TUNSIFHEAD ioctl, and some of them
+ * (e.g. FreeBSD) _need_ it otherwise they'll interpret IPv6 packets as IPv4.
+ */
+#if defined(__OpenBSD__) || defined(TUNSIFHEAD)
 #define TUN_HAS_AF_PREFIX 1
 #endif
 
+#ifdef __sun__
+static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
+{
+	if (!mtu_only)
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "No vpnc-script configured. Need Solaris IP-setting code\n");
+	return 0;
+}
+#else
 static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 {
 	struct ifreq ifr;
@@ -59,7 +93,7 @@ static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 	strncpy(ifr.ifr_name, vpninfo->ifname, sizeof(ifr.ifr_name) - 1);
 
 	if (!mtu_only) {
-		struct sockaddr_in *addr = (struct sockaddr_in *) &ifr.ifr_addr;
+		struct sockaddr_in addr;
 
 		if (ioctl(net_fd, SIOCGIFFLAGS, &ifr) < 0)
 			perror("SIOCGIFFLAGS");
@@ -68,8 +102,9 @@ static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 		if (ioctl(net_fd, SIOCSIFFLAGS, &ifr) < 0)
 			perror("SIOCSIFFLAGS");
 
-		addr->sin_family = AF_INET;
-		addr->sin_addr.s_addr = inet_addr(vpninfo->vpn_addr);
+		addr.sin_family = AF_INET;
+		addr.sin_addr.s_addr = inet_addr(vpninfo->vpn_addr);
+		memcpy(&ifr.ifr_addr, &addr, sizeof(addr));
 		if (ioctl(net_fd, SIOCSIFADDR, &ifr) < 0)
 			perror("SIOCSIFADDR");
 	}
@@ -82,6 +117,7 @@ static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 
 	return 0;
 }
+#endif
 
 static int setenv_int(const char *opt, int value)
 {
@@ -90,11 +126,22 @@ static int setenv_int(const char *opt, int value)
 	return setenv(opt, buf, 1);
 }
 
+static int netmasklen(struct in_addr addr)
+{
+	int masklen;
+
+	for (masklen = 0; masklen < 32; masklen++) {
+		if (ntohl(addr.s_addr) >= (0xffffffff << masklen))
+			break;
+	}
+	return 32 - masklen;
+}
+
 static int process_split_xxclude(struct openconnect_info *vpninfo,
-				 char *in_ex, char *route, int *nr_incs)
+				 char *in_ex, char *route, int *v4_incs,
+				 int *v6_incs)
 {
 	struct in_addr addr;
-	int masklen;
 	char envname[80];
 	char *slash;
 
@@ -108,13 +155,27 @@ static int process_split_xxclude(struct openconnect_info *vpninfo,
 	}
 
 	*slash = 0;
+
+	if (strchr(route, ':')) {
+		snprintf(envname, 79, "CISCO_IPV6_SPLIT_%sC_%d_ADDR", in_ex,
+			 *v6_incs);
+		setenv(envname, route, 1);
+
+		snprintf(envname, 79, "CISCO_IPV6_SPLIT_%sC_%d_MASKLEN", in_ex,
+			 *v6_incs);
+		setenv(envname, slash+1, 1);
+
+		(*v6_incs)++;
+		return 0;
+	}
+		
 	if (!inet_aton(route, &addr)) {
 		*slash = '/';
 		goto badinc;
 	}
 
 	envname[79] = 0;
-	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_ADDR", in_ex, *nr_incs);
+	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_ADDR", in_ex, *v4_incs);
 	setenv(envname, route, 1);
 
 	/* Put it back how we found it */
@@ -123,19 +184,13 @@ static int process_split_xxclude(struct openconnect_info *vpninfo,
 	if (!inet_aton(slash+1, &addr))
 		goto badinc;
 
-	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASK", in_ex, *nr_incs);
+	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASK", in_ex, *v4_incs);
 	setenv(envname, slash+1, 1);
 
-	for (masklen = 0; masklen < 32; masklen++) {
-		if (ntohl(addr.s_addr) >= (0xffffffff << masklen))
-			break;
-	}
-	masklen = 32 - masklen;
+	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASKLEN", in_ex, *v4_incs);
+	setenv_int(envname, netmasklen(addr));
 
-	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASKLEN", in_ex, *nr_incs);
-	setenv_int(envname, masklen);
-
-	(*nr_incs)++;
+	(*v4_incs)++;
 	return 0;
 }
 
@@ -190,8 +245,29 @@ static void set_script_env(struct openconnect_info *vpninfo)
 
 	setenv_int("INTERNAL_IP4_MTU", vpninfo->mtu);
 
-	setenv("INTERNAL_IP4_ADDRESS", vpninfo->vpn_addr, 1);
-	setenv("INTERNAL_IP4_NETMASK", vpninfo->vpn_netmask, 1);
+	if (vpninfo->vpn_addr) {
+		setenv("INTERNAL_IP4_ADDRESS", vpninfo->vpn_addr, 1);
+		if (vpninfo->vpn_netmask) {
+			struct in_addr addr;
+			struct in_addr mask;
+
+			if (inet_aton(vpninfo->vpn_addr, &addr) &&
+			    inet_aton(vpninfo->vpn_netmask, &mask)) {
+				char *netaddr;
+
+				addr.s_addr &= mask.s_addr;
+				netaddr = inet_ntoa(addr);
+
+				setenv("INTERNAL_IP4_NETADDR", netaddr, 1);
+				setenv("INTERNAL_IP4_NETMASK", vpninfo->vpn_netmask, 1);
+				setenv_int("INTERNAL_IP4_NETMASKLEN", netmasklen(mask));
+			}
+		}
+	}
+	if (vpninfo->vpn_addr6) {
+		setenv("INTERNAL_IP6_ADDRESS", vpninfo->vpn_addr6, 1);
+		setenv("INTERNAL_IP6_NETMASK", vpninfo->vpn_netmask6, 1);
+	}
 
 	if (vpninfo->vpn_dns[0])
 		setenv("INTERNAL_IP4_DNS", vpninfo->vpn_dns[0], 1);
@@ -221,32 +297,43 @@ static void set_script_env(struct openconnect_info *vpninfo)
 	if (vpninfo->split_includes) {
 		struct split_include *this = vpninfo->split_includes;
 		int nr_split_includes = 0;
+		int nr_v6_split_includes = 0;
 
 		while (this) {
 			process_split_xxclude(vpninfo, "IN", this->route,
-					      &nr_split_includes);
+					      &nr_split_includes,
+					      &nr_v6_split_includes);
 			this = this->next;
 		}
-		setenv_int("CISCO_SPLIT_INC", nr_split_includes);
+		if (nr_split_includes)
+			setenv_int("CISCO_SPLIT_INC", nr_split_includes);
+		if (nr_v6_split_includes)
+			setenv_int("CISCO_IPV6_SPLIT_INC", nr_v6_split_includes);
 	}
 	if (vpninfo->split_excludes) {
 		struct split_include *this = vpninfo->split_excludes;
 		int nr_split_excludes = 0;
+		int nr_v6_split_excludes = 0;
 
 		while (this) {
 			process_split_xxclude(vpninfo, "EX", this->route,
-					      &nr_split_excludes);
+					      &nr_split_excludes,
+					      &nr_v6_split_excludes);
 			this = this->next;
 		}
-		setenv_int("CISCO_SPLIT_EXC", nr_split_excludes);
+		if (nr_split_excludes)
+			setenv_int("CISCO_SPLIT_EXC", nr_split_excludes);
+		if (nr_v6_split_excludes)
+			setenv_int("CISCO_IPV6_SPLIT_EXC", nr_v6_split_excludes);
 	}
 	setenv_cstp_opts(vpninfo);
 }
 
 static int script_config_tun(struct openconnect_info *vpninfo)
 {
-	if (vpninfo->peer_addr->sa_family != AF_INET) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Script cannot handle anything but Legacy IP\n");
+	if (vpninfo->peer_addr->sa_family != AF_INET || !vpninfo->vpn_addr) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Script can only handle Legacy IP\n");
 		return -EINVAL;
 	}
 
@@ -309,6 +396,81 @@ int setup_tun(struct openconnect_info *vpninfo)
 		}
 		if (!vpninfo->ifname)
 			vpninfo->ifname = strdup(ifr.ifr_name);
+#elif defined (__sun__)
+		static char tun_name[80];
+		int tun2_fd, ip_fd = open("/dev/ip", O_RDWR);
+		int unit_nr, mux_id;
+		struct ifreq ifr;
+
+		if (ip_fd < 0) {
+			perror("open /dev/ip");
+			return -EIO;
+		}
+
+		tun_fd = open("/dev/tun", O_RDWR);
+		if (tun_fd < 0) {
+			perror("open /dev/tun");
+			close(ip_fd);
+			return -EIO;
+		}
+
+		unit_nr = ioctl(tun_fd, TUNNEWPPA, -1);
+		if (unit_nr < 0) {
+			perror("Failed to create new tun");
+			close(tun_fd);
+			close(ip_fd);
+			return -EIO;
+		}
+		
+		tun2_fd = open("/dev/tun", O_RDWR);
+		if (tun2_fd < 0) {
+			perror("open /dev/tun again");
+			close(tun_fd);
+			close(ip_fd);
+			return -EIO;
+		}
+		if (ioctl(tun2_fd, I_PUSH, "ip") < 0) {
+			perror("Can't push IP");
+			close(tun2_fd);
+			close(tun_fd);
+			close(ip_fd);
+			return -EIO;
+		}
+		if (ioctl(tun2_fd, IF_UNITSEL, &unit_nr) < 0) {
+			perror("Can't select unit");
+			close(tun2_fd);
+			close(tun_fd);
+			close(ip_fd);
+			return -EIO;
+		}
+		mux_id = ioctl(ip_fd, I_PLINK, tun2_fd);
+		if (mux_id < 0) {
+			perror("Can't link tun to IP");
+			close(tun2_fd);
+			close(tun_fd);
+			close(ip_fd);
+			return -EIO;
+		}
+		close(tun2_fd);
+
+		sprintf(tun_name, "tun%d", unit_nr);
+		vpninfo->ifname = tun_name;
+
+		memset(&ifr, 0, sizeof(ifr));
+		strcpy(ifr.ifr_name, tun_name);
+		ifr.ifr_ip_muxid = mux_id;
+
+		if (ioctl(ip_fd, SIOCSIFMUXID, &ifr) < 0) {
+			perror("Set mux id");
+			close(tun_fd);
+			ioctl(ip_fd, I_PUNLINK, mux_id);
+			close(ip_fd);
+			return -EIO;
+		}
+		/* Solaris tunctl needs this in order to tear it down */
+		vpninfo->progress(vpninfo, PRG_DEBUG, "mux id is %d\n", mux_id);
+		vpninfo->tun_muxid = mux_id;
+		vpninfo->ip_fd = ip_fd;
 
 #else /* BSD et al have /dev/tun$x devices */
 		static char tun_name[80];
@@ -324,6 +486,13 @@ int setup_tun(struct openconnect_info *vpninfo)
 			exit(1);
 		}
 		vpninfo->ifname = tun_name + 5;
+#ifdef TUNSIFHEAD
+		i = 1;
+		if (ioctl(tun_fd, TUNSIFHEAD, &i) < 0) {
+			perror("TUNSIFHEAD");
+			exit(1);
+		}
+#endif
 #endif
 		if (vpninfo->vpnc_script) {
 			script_config_tun(vpninfo);
@@ -356,24 +525,12 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	if (FD_ISSET(vpninfo->tun_fd, &vpninfo->select_rfds)) {
 		while ((len = read(vpninfo->tun_fd, buf, sizeof(buf))) > 0) {
 			unsigned char *pkt = buf;
-			int type;
-#ifdef __OpenBSD__
-			type = ntohl(*(int *)buf);
-			if (type != AF_INET && type != AF_INET6)
-				continue;
+#ifdef TUN_HAS_AF_PREFIX
 			pkt += 4;
 			len -= 4;
-#else
-			struct ip *iph = (void *)buf;
-			if (iph->ip_v == 6)
-				type = AF_INET6;
-			else if (iph->ip_v == 4)
-				type = AF_INET;
-			else
-				continue;
 #endif
-			if (queue_new_packet(&vpninfo->outgoing_queue, type,
-					     pkt, len))
+			if (queue_new_packet(&vpninfo->outgoing_queue, pkt,
+					     len))
 				break;
 
 			work_done = 1;
@@ -395,13 +552,30 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		unsigned char *data = this->data;
 		int len = this->len;
 
-		vpninfo->incoming_queue = this->next;
+#ifdef TUN_HAS_AF_PREFIX
+		struct ip *iph = (void *)data;
+		int type;
 
-#ifdef __OpenBSD__
+		if (iph->ip_v == 6)
+			type = AF_INET6;
+		else if (iph->ip_v == 4)
+			type = AF_INET;
+		else {
+			static int complained = 0;
+			if (!complained) {
+				complained = 1;
+				vpninfo->progress(vpninfo, PRG_ERR,
+						  "Unknown packet (len %d) received: %02x %02x %02x %02x...\n",
+						  len, data[0], data[1], data[2], data[3]);
+			}
+			free(this);
+			continue;
+		}
 		data -= 4;
 		len += 4;
-		*(int *)data = htonl(this->type);
+		*(int *)data = htonl(type);
 #endif
+		vpninfo->incoming_queue = this->next;
 
 		if (write(vpninfo->tun_fd, data, len) < 0 &&
 		    errno == ENOTCONN) {
@@ -413,3 +587,26 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	/* Work is not done if we just got rid of packets off the queue */
 	return work_done;
 }
+
+void shutdown_tun(struct openconnect_info *vpninfo)
+{	
+	if (vpninfo->script_tun) {
+		kill(vpninfo->script_tun, SIGHUP);
+	} else {
+		if (vpninfo->vpnc_script) {
+			setenv("TUNDEV", vpninfo->ifname, 1);
+			setenv("reason", "disconnect", 1);
+			system(vpninfo->vpnc_script);
+		}
+#ifdef __sun__
+		if (ioctl(vpninfo->ip_fd, I_PUNLINK, vpninfo->tun_muxid) < 0)
+			perror("ioctl(I_PUNLINK)");
+
+		close(vpninfo->ip_fd);
+		vpninfo->ip_fd = -1;
+#endif
+	}
+
+	close(vpninfo->tun_fd);
+	vpninfo->tun_fd = -1;
+}
